@@ -22,10 +22,12 @@ import numpy as np
 from numpy.compat import asbytes, asstr, asunicode
 
 from nums.core import settings
-from nums.core.storage.storage import ArrayGrid, StoredArrayS3
-from nums.core.storage import utils as storage_utils
-from nums.core.systems.systems import System
 from nums.core.array.blockarray import BlockArray
+from nums.core.compute.compute_manager import ComputeManager
+from nums.core.grid.grid import ArrayGrid
+from nums.core.grid.grid import DeviceID
+from nums.core.storage import utils as storage_utils
+from nums.core.storage.storage import StoredArrayS3
 
 
 ################
@@ -155,7 +157,6 @@ def loadtxt_block(fname, dtype, comments, delimiter,
 # CSV API
 ##############
 def read_csv_block(filename, file_start, file_end, dtype, delimiter, has_header):
-
     def _getconv(_dtype):
         """ Adapted from numpy/lib/npyio.py """
 
@@ -221,128 +222,113 @@ class FileSystem(object):
     #  - Less stringent replication of meta-data.
     #  - Journaling?
 
-    def __init__(self, system: System):
-        self.system = system
+    def __init__(self, cm: ComputeManager):
+        self.cm: ComputeManager = cm
         for func in [write_meta_s3, delete_meta_s3,
                      write_block_s3, read_block_s3, delete_block_s3,
                      write_meta_fs, read_meta_fs, delete_meta_fs,
                      write_block_fs, read_block_fs, delete_block_fs,
                      loadtxt_block, read_csv_block]:
-            self.system.register(func.__name__, func, {})
+            self.cm.register(func.__name__, func, {})
 
     ##################################################
     # Block-level (remote) operations
     ##################################################
 
     def write_meta_s3(self, filename: AnyStr, grid_meta: Dict, syskwargs: Dict):
-        return self.system.call("write_meta_s3",
-                                filename,
-                                grid_meta,
-                                syskwargs=syskwargs)
+        return self.cm.call("write_meta_s3",
+                            filename,
+                            grid_meta,
+                            syskwargs=syskwargs)
 
     def delete_meta_s3(self, filename: AnyStr, syskwargs: Dict):
-        return self.system.call("delete_meta_s3", filename, syskwargs=syskwargs)
+        return self.cm.call("delete_meta_s3", filename, syskwargs=syskwargs)
 
     def write_block_s3(self, block: Any, filename: AnyStr, grid_entry: Tuple, grid_meta: Dict,
                        syskwargs: Dict):
-        return self.system.call("write_block_s3",
-                                block,
-                                filename,
-                                grid_entry,
-                                grid_meta,
-                                syskwargs=syskwargs)
+        return self.cm.call("write_block_s3",
+                            block,
+                            filename,
+                            grid_entry,
+                            grid_meta,
+                            syskwargs=syskwargs)
 
     def read_block_s3(self, filename: AnyStr, grid_entry: Tuple, grid_meta: Dict,
                       syskwargs: Dict):
-        return self.system.call("read_block_s3",
-                                filename,
-                                grid_entry,
-                                grid_meta,
-                                syskwargs=syskwargs)
+        return self.cm.call("read_block_s3",
+                            filename,
+                            grid_entry,
+                            grid_meta,
+                            syskwargs=syskwargs)
 
     def delete_block_s3(self, filename: AnyStr, grid_entry: Tuple, grid_meta: Dict,
                         syskwargs: Dict):
-        return self.system.call("delete_block_s3",
-                                filename,
-                                grid_entry,
-                                grid_meta,
-                                syskwargs=syskwargs)
+        return self.cm.call("delete_block_s3",
+                            filename,
+                            grid_entry,
+                            grid_meta,
+                            syskwargs=syskwargs)
 
     def loadtxt_block(self, fname, dtype, comments, delimiter,
                       converters, skiprows, usecols, unpack,
                       ndmin, encoding, max_rows, syskwargs: Dict):
         # TODO (hme): Invoke file_exists with options to determine which nodes to pull from.
-        return self.system.call("loadtxt_block",
-                                fname, dtype, comments, delimiter,
-                                converters, skiprows, usecols, unpack,
-                                ndmin, encoding, max_rows,
-                                syskwargs=syskwargs)
+        return self.cm.call("loadtxt_block",
+                            fname, dtype, comments, delimiter,
+                            converters, skiprows, usecols, unpack,
+                            ndmin, encoding, max_rows,
+                            syskwargs=syskwargs)
 
     def write_block_fs(self, block: Any, filename: AnyStr, grid_entry: Tuple, grid_meta: Dict,
                        syskwargs: Dict):
-        return self.system.call("write_block_fs",
-                                block,
-                                filename,
-                                grid_entry,
-                                syskwargs=syskwargs)
+        return self.cm.call("write_block_fs", block, filename, grid_entry,
+                            syskwargs=syskwargs)
 
     def read_block_fs(self, filename: AnyStr, grid_entry: Tuple, grid_meta: Dict,
-                      options: Dict):
-        return self.system.call_with_options("read_block_fs",
-                                             args=[filename, grid_entry],
-                                             kwargs={},
-                                             options=options)
+                      syskwargs: Dict):
+        return self.cm.call("read_block_fs", filename, grid_entry,
+                            syskwargs=syskwargs)
 
     def delete_block_fs(self, filename: AnyStr, grid_entry: Tuple, grid_meta: Dict,
-                        options: Dict):
-        return self.system.call_with_options("delete_block_fs",
-                                             args=[filename, grid_entry],
-                                             kwargs={},
-                                             options=options)
+                        syskwargs: Dict):
+        return self.cm.call("delete_block_fs", filename, grid_entry,
+                            syskwargs=syskwargs)
 
     ##################################################
     # Array-level operations
     ##################################################
 
     def write_meta_fs(self, ba: BlockArray, filename: str):
-        addresses: dict = self.system.get_block_addresses(ba.grid)
+        addresses: dict = {}
+        for grid_entry in ba.grid.get_entry_iterator():
+            device_id: DeviceID = self.cm.device_grid.get_device_id(grid_entry, ba.grid.grid_shape)
+            addresses[grid_entry] = str(device_id)
         meta = {"filename": filename,
                 "grid_meta": ba.grid.to_meta(),
                 "addresses": addresses}
         oids = []
         for grid_entry in ba.grid.get_entry_iterator():
-            node_name = addresses[grid_entry]
-            oid = self.system.call_with_options("write_meta_fs",
-                                                args=[meta, filename],
-                                                kwargs={},
-                                                options={"resources": {node_name: 1.0/10**4}})
+            device_id: DeviceID = self.cm.device_grid.get_device_id(grid_entry,
+                                                                    ba.grid.grid_shape)
+            oid = self.cm.call("write_meta_fs",
+                               meta,
+                               filename,
+                               syskwargs={"device_id": device_id})
             oids.append(oid)
         return oids
 
     def read_meta_fs(self, filename: str):
-        for node in self.system.nodes():
-            node_key = list(filter(lambda key: "node" in key, node["Resources"].keys()))
-            assert len(node_key) == 1
-            node_name = node_key[0]
-            oid = self.system.call_with_options("read_meta_fs",
-                                                args=[filename],
-                                                kwargs={},
-                                                options={"resources": {node_name: 1.0/10**4}})
-            result = self.system.get(oid)
+        for device_id in self.cm.devices():
+            oid = self.cm.call("read_meta_fs", filename, syskwargs={"device_id": device_id})
+            result = self.cm.get(oid)
             if result is not None:
                 return result
         raise Exception("failed to load metadata.")
 
     def delete_meta_fs(self, filename: str):
         oids = []
-        for node in self.system.nodes():
-            node_key = list(filter(lambda key: "node" in key, node["Resources"].keys()))
-            assert len(node_key) == 1
-            node_name = node_key[0]
-            oid = self.system.call_with_options("delete_meta_fs",
-                                                args=[filename],
-                                                kwargs={},
-                                                options={"resources": {node_name: 1.0/10**4}})
+        for device_id in self.cm.devices():
+            oid = self.cm.call("delete_meta_fs", filename, syskwargs={"device_id": device_id})
             oids.append(oid)
         return oids
 
@@ -381,10 +367,10 @@ class FileSystem(object):
         grid = ArrayGrid(shape=(num_rows_final, num_cols),
                          block_shape=(row_batches.batch_size, num_cols),
                          dtype=np.float64.__name__ if dtype is float else dtype.__name__)
-        result: BlockArray = BlockArray(grid, system=self.system)
+        result: BlockArray = BlockArray(grid, cm=self.cm)
         for i, grid_entry in enumerate(grid.get_entry_iterator()):
             row_start, row_end = row_batches.batches[i]
-            batch_skiprows = skiprows + row_start + 1
+            batch_skiprows = skiprows + row_start
             batch_max_rows = grid.get_block_shape(grid_entry)[0]
             assert batch_max_rows == row_end - row_start
             result.blocks[grid_entry].oid = self.loadtxt_block(
@@ -399,7 +385,7 @@ class FileSystem(object):
             )
         return result
 
-    def read_csv(self, filename, dtype=np.float, delimiter=',', has_header=False, num_workers=4):
+    def read_csv(self, filename, dtype=float, delimiter=',', has_header=False, num_workers=4):
         file_size = storage_utils.get_file_size(filename)
         file_batches: storage_utils.Batch = storage_utils.Batch.from_num_batches(file_size,
                                                                                  num_workers)
@@ -407,21 +393,21 @@ class FileSystem(object):
         shape_oids = []
         for i, batch in enumerate(file_batches.batches):
             file_start, file_end = batch
-            block_oid, shape_oid = self.system.call("read_csv_block",
-                                                    filename,
-                                                    file_start,
-                                                    file_end,
-                                                    dtype,
-                                                    delimiter,
-                                                    has_header,
-                                                    syskwargs={
-                                                        "grid_entry": (i,),
-                                                        "grid_shape": (num_workers,),
-                                                        "options": {"num_returns": 2}
-                                                    })
+            block_oid, shape_oid = self.cm.call("read_csv_block",
+                                                filename,
+                                                file_start,
+                                                file_end,
+                                                dtype,
+                                                delimiter,
+                                                has_header,
+                                                syskwargs={
+                                                    "grid_entry": (i,),
+                                                    "grid_shape": (num_workers,),
+                                                    "options": {"num_returns": 2}
+                                                })
             blocks.append(block_oid)
             shape_oids.append(shape_oid)
-        shapes = self.system.get(shape_oids)
+        shapes = self.cm.get(shape_oids)
         arrays = []
         for i in range(len(shapes)):
             shape = shapes[i]
@@ -429,7 +415,7 @@ class FileSystem(object):
                 continue
             block = blocks[i]
             grid = ArrayGrid(shape=shape, block_shape=shape, dtype=dtype.__name__)
-            arr = BlockArray(grid, self.system)
+            arr = BlockArray(grid, self.cm)
             iter_one = True
             for grid_entry in grid.get_entry_iterator():
                 assert iter_one
